#####################################################################################################
#############################   Comparing Bayesian variable selection methods #################################
###################################################################################################
################################# Sections 5.2 and A.5 ##################################################################
###################################################################################################


library(mvtnorm)
library(boot)
library(entropy)
require(graphics)
library(tictoc)
library(matlib)
library(EMVS)
library(spikeslab)
library(varbvs)
library(selectiveInference)
require(MASS)
library(SSLASSO)
library(horseshoe)
library(lars)
library(matrixStats)
library(sparsevb)




# Experiment 1:
n<-100
p<-n*4
s<-10
value<-runif(s,-3,3)
theta<-rep(0,p)
theta[1:s]<- value
zeta<-0.2
#rho<-0.3

# Experiment 2 (as experiment 1, but)
#rho<-0.7


# Experiment 3:
n<-200
p<-n*4
s<-20
value<-2*log(n)
theta<-rep(0,p)
theta[(p-s+1):p]<- value
zeta<-5
#rho<-0.3

# Experiment 4 (as experiment 3, but):
 rho<-0.7


lambda_number<-200 # number of cycles
v_number<-2

repeat_number<-5


errorSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )


timeL<-rep(0,repeat_number)
errorL<-rep(0,repeat_number)
fdrL<-rep(0,repeat_number)
tprL<-rep(0,repeat_number)
timevar<-rep(0,repeat_number)
errorvar<-rep(0,repeat_number)
fdrvar<-rep(0,repeat_number)
tprvar<-rep(0,repeat_number)
timeEMVS<-rep(0,repeat_number)
errorEMVS<-matrix(rep(0,repeat_number*v_number),nrow=repeat_number )
fdrEMVS<-matrix(rep(0,repeat_number*v_number),nrow=repeat_number )
tprEMVS<-matrix(rep(0,repeat_number*v_number),nrow=repeat_number )
timeSSL<-rep(0,repeat_number)
errorSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
fdrSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
tprSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
timesSSL<-rep(0,repeat_number)
timeEB<-rep(0,repeat_number)
errorEB<-rep(0,repeat_number)
fdrEB<-rep(0,repeat_number)
tprEB<-rep(0,repeat_number)

for (j in 1:repeat_number){
print(j)
  # Generating the data and the design matrix:
  eps   =  rnorm(n,0,1)  
  Sigma<-matrix(rep(rho,p*p), nrow=p)
  diag(Sigma)<-1
  X<- mvrnorm(n = n, mu=rep(0,p), Sigma, tol = 1e-6, empirical = FALSE)
  Y =  X %*% theta +eps*zeta

  
  
  ################################ VB Laplace ##################
  #initialization 
  tic()
  # estimating sigma empirically 
  hatsigma<-0
  while(hatsigma<0.01|hatsigma>100 | hatsigma=="NaN"){
    hatsigma<-estimateSigma(x=X,y=Y)$sigmahat
  }
  Xrescale<-X/hatsigma
  Yrescale<-Y/hatsigma
  vbL<-svb.fit(Xrescale,Yrescale,tol = 10e-4, max_iter = 10^4, exact_math = TRUE, family ="linear")
  mu<-vbL$mu
  sigma<-vbL$sigma
  gamma<-vbL$gamma
  
  time0<-toc()
  timeL[j]<-time0$toc-time0$tic
  errorL[j]<-sqrt(sum((theta-t(mu*gamma))^2))
  fdrL[j]<-sum((theta==0)*(gamma>0.5))/max(sum(gamma>0.5),1)
  tprL[j]<-sum((theta!=0)*(gamma>0.5))/max(sum(theta!=0),1)
  
  ########################### varbvs ###########################
  
  tic()
  #initialisation:
  XTX= t(X) %*% X
  YTX<-t(X)%*%Y
  mu0<-solve(XTX+diag(p))%*%YTX
  
  help<-varbvs(X=X, Z=NULL,y=Y, family = "gaussian",  mu=mu0 ,verbose = TRUE, tol = 1e-4, maxiter = 10000)
  thetaM<-help$beta
  pip<-help$pip
  time0<-toc()
  timevar[j]<-time0$toc-time0$tic
  errorvar[j]<-sqrt(sum((theta-t(thetaM))^2))
  fdrvar[j]<-sum((theta==0)*(pip>0.5))/max(1,sum(pip>0.5))
  tprvar[j]<-sum((theta!=0)*(pip>0.5))/sum(theta!=0)
  
  
  
  ########################### EMVS ###########################
  tic()
  # initialisation
  v0 = seq(0.1, 2, length = v_number)
  v1 = 1000
  sigma_init = 1
  a = 1
  b = p
  epsilon = 10^{-5}
  beta_init<-solve(XTX+diag(p))%*%YTX
  # calling the EMVS routine
  obj<-EMVS(Y, X,v0=v0, v1=v1,type = c("betabinomial"), independent = TRUE,
       beta_init, epsilon=epsilon,a=a,b=b)
  
  thetaEMVS<-obj$betas
  pipEMVS<-obj$prob_inclusion
  time0<-toc()
  timeEMVS[j]<-time0$toc-time0$tic
  
  for (i in 1:v_number){
    errorEMVS[j,i]<-sqrt(sum((theta-thetaEMVS[i,])^2))
    fdrEMVS[j,i]<-sum((theta==0)*(pipEMVS[i,]>0.5))/max(sum(pipEMVS[i,]>0.5),1)
    tprEMVS[j,i]<-sum((theta!=0)*(pipEMVS[i,]>0.5))/sum(theta!=0)
  } 
  
  
  ########################### SSLASSO ###########################
  tic()
  ## initialisation:
  lambda1<-0.01
  lambda0<-seq(lambda1,p,length=lambda_number)
  ## calling the SSLASSO routine
  obj<-SSLASSO(X, Y, variance = "unknown",lambda1 = lambda1, lambda0 = lambda0,a=1,b=p)
  thetaSSL<-obj$beta
  pipSSL<-obj$select
  time0<-toc()
  timeSSL[j]<-time0$toc-time0$tic
  
  for (i in 1:lambda_number){
    errorSSL[j,i]<-sqrt(sum((theta-thetaSSL[,i])^2))
    fdrSSL[j,i]<-sum((theta==0)*(pipSSL[,i]>0.5))/max(sum(pipSSL[,i]>0.5),1)
    tprSSL[j,i]<-sum((theta!=0)*(pipSSL[,i]>0.5))/sum(theta!=0)
  } 
  
  
  
  
  ########################### Empirical Bayes - R function ###########################
  
  
  #tic()
  ## initialisation
  #M<-5000
  #sample.beta <- TRUE
  #alpha <- 0.99
  #gam <- 0.001
  #log.f <- function(x) dcomplex(x, n, p, 0.05, 1)
  #hatsigma<-0
  ## estimating the variance
  #while(hatsigma<0.01|hatsigma>100 | hatsigma=="NaN"){
  #  hatsigma<-estimateSigma(x=X,y=Y)$sigmahat
  #}
  ## calling the ebreg.R function
  #o <- ebreg2(Y, X, sig2=hatsigma, log.f, alpha, gam, M, sample.beta)
  #sample_beta<-o$beta
  #sample_S<-o$S
  #postmean_MCMC<-colMeans(sample_beta, na.rm = FALSE, dims = 1)
  #pipEB<-colMeans(sample_S, na.rm = FALSE, dims = 1)
  #time0<-toc()
  
  #timeEB[j]<-time0$toc-time0$tic
  #errorEB[j]<-sqrt(sum((theta- postmean_MCMC )^2)) 
  #fdrEB[j]<-sum((theta==0)*(pipEB>0.5))/max(sum(pipEB>0.5),1)
  #tprEB[j]<-sum((theta!=0)*(pipEB>0.5))/sum(theta!=0)
  
  
  
   
}


# Computing the mean and standard deviation of the above indicatior over 'repeat_number' runs:


errorLavg<-mean(errorL)
errorLsd<-sd(errorL)
errorvaravg<-mean(errorvar)
errorvarsd<-sd(errorvar)
errorEMVSavg<-colMeans(errorEMVS)
errorEMVSsd<-colSds(errorEMVS)
error_meanSSL<-colMeans(errorSSL)
error_sdSSL<-colSds(errorSSL)
errorEBavg<-mean(errorEB)
errorEBsd<-sd(errorEB)

errorLavg
errorLsd
errorvaravg
errorvarsd
errorEMVSavg
errorEMVSsd
error_meanSSL
error_sdSSL
errorEBavg
errorEBsd


fdrLavg<-mean(fdrL)
fdrLsd<-sd(fdrL)
fdrvaravg<-mean(fdrvar)
fdrvarsd<-sd(fdrvar)
fdrEMVSavg<-colMeans(fdrEMVS)
fdrEMVSsd<-colSds(fdrEMVS)
fdrSSLavg<-colMeans(fdrSSL)
fdrSSLsd<-colSds(fdrSSL)
fdrEBavg<-mean(fdrEB)
fdrEBsd<-sd(fdrEB)

fdrLavg
fdrLsd
fdrvaravg
fdrvarsd
fdrEMVSavg
fdrEMVSsd
fdrSSLavg
fdrSSLsd
fdrEBavg
fdrEBsd

tprLavg<-mean(tprL)
tprLsd<-sd(tprL)
tprvaravg<-mean(tprvar)
tprvarsd<-sd(tprvar)
tprEMVSavg<-colMeans(tprEMVS)
tprEMVSsd<-colSds(tprEMVS)
tprSSLavg<-colMeans(tprSSL)
tprSSLsd<-colSds(tprSSL)
tprEBavg<-mean(tprEB)
tprEBsd<-sd(tprEB)

tprLavg
tprLsd
tprvaravg
tprvarsd
tprEMVSavg
tprEMVSsd
tprSSLavg
tprSSLsd
tprEBavg
tprEBsd

timeLavg<-mean(timeL)
timeLsd<-sd(timeL)
timevaravg<-mean(timevar)
timevarsd<-sd(timevar)
timeEMVSavg<-mean(timeEMVS)
timeEMVSsd<-sd(timeEMVS)
timeSSLavg<-mean(timeSSL)
timeSSLsd<-sd(timeSSL)
timeEBavg<-mean(timeEB)
timeEBsd<-sd(timeEB)

timeLavg
timeLsd
timevaravg
timevarsd
timeEMVSavg
timeEMVSsd
timeSSLavg
timeSSLsd
timeEBavg
timeEBsd





########################### Pictures - for the last run ###########################

plot(c(1:p),theta , col=3, ylab="Signal value", xlab="Index",  ylim=c(-1*max(theta),2*max(theta)), pch=20,cex.lab=1.5, cex.axis=1.5, cex.main=1.5, cex.sub=1.5)

points(c(1:p),theta,col=3,pch=20)
points(c(1:p),mu*gamma,col=4,pch=20)
points(c(1:p),thetaM, col=2,pch=20)
points(c(1:p),thetaEMVS, col=7,pch=20)
points(c(1:p),thetaSSL, col=5,pch=20)
points(c(1:p),postmean_MCMC, col=8,pch=20)
#legend("topright", col=c(3,4,2,6,7,5,8),cex=1.2, lwd=2, legend=c("signal", "vb - Laplace", "varbvs", "spikeslab","EMVS","SSLASSO", "Empirical Bayes"))


















ebreg2 <- function(y, X, sig2=NULL, log.f, alpha, gam, M, sample.beta=FALSE) {
  
  n <- nrow(X)
  p <- ncol(X)
  if(!exists("lars")) library(lars)
  o.lars <- lars(X, y, normalize=FALSE, intercept=FALSE, use.Gram=FALSE)
  cv <- cv.lars(X, y, plot.it=FALSE, se=FALSE, normalize=FALSE, intercept=FALSE, use.Gram=FALSE)
  b.lasso <- coef(o.lars, s=cv$index[which.min(cv$cv)], mode="fraction")
  S <- as.numeric(b.lasso != 0)
  if(is.null(sig2)) {
    
    z <- as.numeric(y - X[, S > 0] %*% b.lasso[S > 0])
    sig2 <- sum(z**2) / max(n - sum(S), 1)
  }
  v <- 1 + alpha / gam / sig2
  sq.v <- sqrt(v)
  B <- round(0.2 * M)
  bb <- if(sample.beta) matrix(0, nrow=B + M, ncol=p) else NULL
  SS <- matrix(0, nrow=B + M, ncol=p)
  ss <- numeric(B + M)
  ii <- integer(B + M)
  SS[1,] <- S
  ss[1] <- s <- sum(S)
  ii[1] <- i <- iuS <- nuS <- 1
  out <- list()
  out[[1]] <- get.lm.stuff(S, y, X, sample.beta)
  lprior <- log.f(0:n) - lchoose(p, 0:n)
  lpost <- lprior[s] - alpha * out[[1]]$sse / 2 / sig2 - s * log(v) / 2
  for(m in 1:(B + M)) {
    
    S.new <- rprop(S, n)
    s.new <- sum(S.new)
    i.new <- compare.to.rows(as.matrix(SS[iuS,]), S.new)
    if(i.new == 0) o.new <- get.lm.stuff(S.new, y, X, sample.beta) else o.new <- out[[i.new]]
    lpost.new <- lprior[s.new] - alpha * o.new$sse / 2 / sig2 - s.new * log(v) / 2
    if(runif(1) <= exp(lpost.new - lpost)) {
      
      S <- S.new
      s <- s.new
      lpost <- lpost.new
      if(i.new == 0) {
        
        nuS <- nuS + 1
        i <- nuS
        iuS <- c(iuS, m)
        out[[nuS]] <- o.new
        
      } else i <- i.new
      
    }
    SS[m,] <- S
    ss[m] <- s
    ii[m] <- i
    if(sample.beta) {
      
      b <- numeric(p)
      b[S > 0] <- out[[i]]$b.hat + sq.v * out[[i]]$U %*% rnorm(s)
      bb[m,] <- b
      
    }
    
  }
  if(sample.beta) bb[-(1:B),]
  return(list(beta=bb, S=SS[-(1:B),], s=ss[-(1:B)], i=ii[-(1:B)]))
  
}



# Auxiliary functions called by 'ebreg'


dcomplex <- function(x, n, p, a, b, log=TRUE) {
  
  o <- -x * (log(b) + a * log(p)) + log(x <= n)
  if(!log) o <- exp(o)
  return(o)
  
}



rprop <- function(S, n) {
  
  s <- sum(S)
  if(s == n) { S[sample(which(S == 1), 1)] <- 0 }
  else if(s  == 1) { S[sample(which(S == 0), 1)] <- 1 }
  else {
    
    if(runif(1) <= 0.5) S[sample(which(S == 1), 1)] <- 0
    else S[sample(which(S == 0), 1)] <- 1
    
  }
  return(S)
  
}


compare.to.rows <- function(SS, S) {
  
  h <- function(v) sum(abs(v - S))
  o <- apply(SS, 1, h)
  if(all(o > 0)) return(0) else return(which(o == 0)[1])
  
}


get.lm.stuff <- function(S, y, X, sample.beta) {
  
  if(!exists("ginv")) library(MASS)
  X.S <- as.matrix(X[, S > 0])
  o <- lm(y ~ X.S - 1)
  sse <- sum(o$residuals**2)
  if(sample.beta) {
    
    b.hat <- o$coefficients
    XtX.S <- t(X.S) %*% X.S
    V <- ginv(XtX.S)
    U <- chol(V)
    
  } else {
    
    U <- NULL
    b.hat <- NULL
    
  }
  return(list(sse=sse, b.hat=b.hat, U=U))
  
}







