#####################################################################################################
###############   Comparing different choices of lambda in our vb method ##############
###################################################################################################
################################# Sections A.3 #####################################################
###################################################################################################




library(mvtnorm) 
library(boot)
library(entropy)
require(graphics)
library(tictoc)
library(sparsevb)



#########################################




################# Initialization  #################


# Experiment 1:
#n<-200
#p<-300
#s<-15
#value=2*log(n)
#tau=0.5
#theta<-rep(0,p)
#theta[1:s]<-value

# Experiment 2:
n<-500
p<-1000
s<-50
value=2*log(n)
tau=1
theta<-rep(0,p)
theta[1:s]<-value

# Experiment 3:
#<-200
#p<-500
#s<-20
#value=runif(s,-10,10)
#tau=0.2
#theta<-rep(0,p)
#theta[1:s]<-value

# Experiment 4:
#n<-1000
#p<-2000
#s<-15
#value=runif(s,-8,8)
#tau=2
#theta<-rep(0,p)
#theta[1:s]<-value

repeat_number<-100
lambdavect=c(1/20,1/4,1,4,20)
lambdalength<-length(lambdavect)
postmean<-matrix(rep(0,lambdalength*p),nrow = lambdalength)






time_vect<-matrix(rep(0,repeat_number*lambdalength),nrow=repeat_number,ncol=lambdalength)
error<-matrix(rep(0,repeat_number),nrow=repeat_number,ncol=lambdalength)
fdr<-matrix(rep(0,repeat_number*lambdalength),nrow=repeat_number,ncol=lambdalength)
tpr<-matrix(rep(0,repeat_number*lambdalength),nrow=repeat_number,ncol=lambdalength)






for(iter in 1:repeat_number){
print(iter)

  ################# design matrix, observations ################
  X<-matrix(rnorm(n*p, mean=0,sd=tau),nrow=n,ncol=p)
  eps   =  rnorm(n,0,1)                               # Noise epsilon
  Y     =  X %*% theta +eps

  
  
  
  
  for(index in 1:lambdalength){
    # Initialisation:
    lambda<-lambdavect[index]

    
    tic()
    
    vbL<-svb.fit(X,Y,lambda=lambda,tol = 10e-4, max_iter = 10^4, exact_math = TRUE, family ="linear")
    mu<-vbL$mu
    sigma<-vbL$sigma
    gamma<-vbL$gamma
    time0<-toc()
    
    sum(gamma>0.5)
    postmean[index,]<-mu*gamma
    error[iter,index]<- sqrt(sum((postmean[index,]-theta)^2))
    time_vect[iter,index]<-time0$toc-time0$tic
    fdr[iter,index]<-sum((theta==0)*(gamma>0.5))/max(sum(gamma>0.5),1)
    tpr[iter,index]<-sum((theta!=0)*(gamma>0.5))/sum(theta!=0)
  }
  

}

# Computing the mean and standard deviations of l2-error, FDr, TPR and runtime over `repeat_number` runs
erroravg<-rep(0,lambdalength)
errorsd<-rep(0,lambdalength)
fdravg<-rep(0,lambdalength)
fdrsd<-rep(0,lambdalength)
tpravg<-rep(0,lambdalength)
tprsd<-rep(0,lambdalength)
timeavg<-rep(0,lambdalength)
timesd<-rep(0,lambdalength)


for(index in 1:lambdalength){
  erroravg[index]<-mean(error[,index])
  errorsd[index]<-sd(error[,index])
  fdravg[index]<-mean(fdr[,index])
  fdrsd[index]<-sd(fdr[,index])
  tpravg[index]<-mean(tpr[,index])
  tprsd[index]<-sd(tpr[,index])
  timeavg[index]<-mean(time_vect[,index])
  timesd[index]<-sd(time_vect[,index])
}

erroravg
errorsd
fdravg
fdrsd
tpravg
tprsd
timeavg
timesd

# Picture corresponding to the last run:

plot(c(1:p),theta , col=3, ylab="Signal value", xlab="Index", ylim=c(min(c(theta)),max(2*c(theta))), pch=15,cex.lab=1.5, cex.axis=1.5, cex.main=1.5, cex.sub=1.5)
points(c(1:p),postmean[1,],col=4,pch=20)
points(c(1:p),postmean[2,],col=2,pch=20)
points(c(1:p),postmean[3,],col=5,pch=20)
points(c(1:p),postmean[4,],col=8,pch=20)
points(c(1:p),postmean[5,],col="orange",pch=20)
legend("topright", col=c(3,4,2,5,8,"orange"),cex=1.2, lwd=2, legend=c("signal","lambda=0.01","lambda=0.1", "lambda=1", "lambda=10", "lambda=100"))


