################################################################################
#######################   Comapring various VB algorithms  #####################
################################################################################
################################# Section A.2 ##################################
################################################################################




library(boot)
library(entropy)
require(graphics)
library(tictoc)
library(sparsevb)

theta<-rep(0,p)

# Experiment 1:
#n<-400
#p<-400
#X<-diag(n)
#s<-40
#value=4*sqrt(log(n))
#theta[1:s]<-value

## For experiments 2-4 the design matrix is defined inside of the loop (we take different Gaussian random designs)

# Experiment 2:
n<-100
p<-200
s<-20
value=runif(s,0,2*log(n))
theta[1:s]<-value
Xsd=1

# Experiment 3:
#n<-200
#p<-800
#s<-40
#value=2*log(n)
#theta[1:s]<-value
#Xsd=0.1

# Experiment 4:
#n<-100
#p<-400
#s<-15
#value=runif(s,-8,8)
#theta[1:s]<-value
#Xsd=0.5

# Initialization:
repeat_number<-10 # number of runs

timeL<-rep(0,repeat_number)
errorL<-rep(0,repeat_number)
fdrL<-rep(0,repeat_number)
tprL<-rep(0,repeat_number)
timeL2<-rep(0,repeat_number)
errorL2<-rep(0,repeat_number)
fdrL2<-rep(0,repeat_number)
tprL2<-rep(0,repeat_number)
timeG<-rep(0,repeat_number)
errorG<-rep(0,repeat_number)
fdrG<-rep(0,repeat_number)
tprG<-rep(0,repeat_number)
timeG2<-rep(0,repeat_number)
errorG2<-rep(0,repeat_number)
fdrG2<-rep(0,repeat_number)
tprG2<-rep(0,repeat_number)
timeG3<-rep(0,repeat_number)
errorG3<-rep(0,repeat_number)
fdrG3<-rep(0,repeat_number)
tprG3<-rep(0,repeat_number)


for (j in 1: repeat_number){

    
  #For experiments: 2-4  
  X<-matrix(rnorm(n*p, mean=0,sd=Xsd),nrow=n,ncol=p)
  eps   =  rnorm(n,0,1)                               # Noise epsilon
  Y     =  X %*% theta + eps
  
  
  
  
  ################################################################################################################
  ################ Component-wise variational Bayes algorithm - Laplace slab - P #############
  ####################################################################################
  
  
  ############################ initialization  ##############
  tic()
  vbL<-svb.fit(X,Y,family ="linear",tol = 10e-4, max_iter = 10^4, exact_math = TRUE )
  muL<-vbL$mu
  sigmL<-vbL$sigma
  gammaL<-vbL$gamma
  a<-toc()
  timeL[j]<-a$toc-a$tic
  errorL[j]<-sum((theta-t(muL*gammaL))^2)
  fdrL[j]<-sum((theta==0)*(gammaL>0.5))/max(1,sum(gammaL>0.5))
  tprL[j]<-sum((theta!=0)*(gammaL>0.5))/sum(theta!=0)
  
  
  
  
  
  
  ################################################################################################################
  ################ Component-wise variational Bayes algorithm - Laplace slab - Q #############
  ####################################################################################
  
  
  
  ############################ initialization  ##############
  tic()
  XTX= t(X) %*% X
  YTX<-t(X)%*%Y
  mu<-solve(XTX+diag(p))%*%YTX
  perm<-order(-abs(mu))
  sigm=rep(1,p)
  lambda<-1
  a0<-1
  b0<-p
  gamma=rep(a0/(a0+b0),p)
  entropy_change<-rep(0,p)
  appr_error<-1
  ################################################ algorithm #############################
  while( appr_error>10^(-4)){
    gamma_old<-gamma
    for(k in 1:p){
      i<-perm[k]
      o<- optimize(fsigmhelp,interval=c(0.01,2),tol = 0.01,maximum = FALSE)
      sigm[i]<-o$minimum
      o<- optimize(fmuhelp,interval=c(-max(abs(mu))*2,max(abs(mu))*2),tol = 0.001,maximum = FALSE)
      mu[i]<- o$minimum
      gamma0<-gamma
      gamma0[i]<-0
      gamma1<-gamma
      gamma1[i]<-1
      
      gamma[i]<-1
      if(fgamma_eval(i, sigm, mu,gamma0,omega)<=fgamma_eval(i, sigm, mu,gamma1,omega) ){
        gamma[i]<-0
      }
      entropy_change[i]<-entropy(c(gamma[i],1-gamma[i]))-entropy(c(gamma_old[i],1-gamma_old[i]))
    }    
    appr_error<-max(abs(entropy_change))
  }
  muL2<-mu
  sigmL2<-sigm
  gammaL2<-gamma
  a<-toc()
  timeL2[j]<-a$toc-a$tic
  errorL2[j]<-sum((theta-t(mu*gamma))^2)
  fdrL2[j]<-sum((theta==0)*(gamma>0.5))/max(1,sum(gamma>0.5))
  tprL2[j]<-sum((theta!=0)*(gamma>0.5))/sum(theta!=0)
  
  
  
  
  
  
  
  
  
  
  
  
  ################################################################################################################
  ################ Component-wise variational Bayes algorithm - Gaussian slab #############
  ####################################################################################
  
  
  
  ############################ initialization  ##############
  tic()
  XTX= t(X) %*% X
  YTX<-t(X)%*%Y
  muG<-solve(XTX+diag(p))%*%YTX
  sigmG=rep(1,p)
  gammaG=rep(a0/(a0+b0),p)
  entropy_changeG<-rep(0,p)
  appr_errorG<-1
  
  ############################ algorithm  ##############
  while(appr_errorG>10^(-3)){
    gamma_oldG<-gammaG
    for(i in 1:p){
      sigmG[i]<- sqrt(1/(XTX[i,i]+1))
      muG[i]<- sigmG[i]^2*(YTX[i]-XTX[-i,i] %*% (gammaG*muG)[-i])
      gammaG[i]<-inv.logit(log(a0/b0)+log(sigmG[i])+muG[i]^2/(2*sigmG[i]^2))
      entropy_changeG[i]<-entropy(c(gammaG[i],1-gammaG[i]))-entropy(c(gamma_oldG[i],1-gamma_oldG[i]))
    }    
    appr_errorG<-max(abs(entropy_changeG))
  }
  a<-toc()
  timeG[j]<-a$toc-a$tic
  errorG[j]<-sum((theta-t(muG*gammaG))^2)
  fdrG[j]<-sum((theta==0)*(gammaG>0.5))/max(1,sum(gammaG>0.5))
  tprG[j]<-sum((theta!=0)*(gammaG>0.5))/sum(theta!=0)
  
  
  
  
  
  ################################################################################################################
  ################ Component-wise variational Bayes algorithm - Gaussian slab -var= ||\theta_0||_2^2 #############
  ####################################################################################
  
  
  ##################### initialization ##########################
  tic()
  XTX= t(X) %*% X
  YTX<-t(X)%*%Y
  muG2<-solve(XTX+diag(p))%*%YTX
  gammaG2=rep(1,p)
  sigmG2=rep(1,p)
  entropy_changeG2<-rep(0,p)
  appr_errorG2<-1
  sigmabeta<-(sum(theta^2))^(1/2)
  ################ algorithm #############
  
  while(appr_errorG2>10^(-3)){
    gamma_oldG2<-gammaG2
    for(i in 1:p){
      sigmG2[i]<- sqrt(1/(XTX[i,i]+1/sigmabeta^2))
      muG2[i]<- sigmG2[i]^2*(YTX[i]-XTX[-i,i] %*% (gammaG2*muG2)[-i])
      gammaG2[i]<-inv.logit(log(a0/b0)+log(sigmG2[i]/sigmabeta)+muG2[i]^2/(2*sigmG2[i]^2))
      entropy_changeG2[i]<-entropy(c(gammaG2[i],1-gammaG2[i]))-entropy(c(gamma_oldG2[i],1-gamma_oldG2[i]))
    }    
    appr_errorG2<-max(abs(entropy_changeG2))
  }
  a<-toc()
  timeG2[j]<-a$toc-a$tic
  errorG2[j]<-sum((theta-t(muG2*gammaG2))^2)
  fdrG2[j]<-sum((theta==0)*(gammaG2>0.5))/max(1,sum(gammaG2>0.5))
  tprG2[j]<-sum((theta!=0)*(gammaG2>0.5))/sum(theta!=0)

  
  ################################################################################################################
  ################  variational Bayes algorithm - Gaussian slab - batch-wise #############
  ####################################################################################
  
  
  ##################### initialization ##########################
  #tic()
  #XTX= t(X) %*% X
  #YTX<-t(X)%*%Y
  #muG3<-solve(XTX+diag(p))%*%YTX
  #gammaG3=rep(1,p)
  #sigmG3=rep(1,p)
  #entropy_changeG3<-rep(0,p)
  #appr_errorG3<-1
  #iterUB<-100
  #iter<-1
  ################ algorithm #############
  
  #while(appr_errorG3>10^(-3) &iter<iterUB){
  #  iter<-iter+1
  #  gamma_oldG3<-gammaG3
  #  sigmG3<- sqrt(1/(diag(XTX)+gammaG3))
  #  for(i in 1:p){gammaG3[i]<-inv.logit(log(1/p)+log(sigmG3[i])+muG3[i]^2/(2*sigmG3[i]^2))}
  #  muG3<-  t(solve(XTX+diag(gammaG3))%*%YTX)
  #  for(i in 1:p){
  #    entropy_changeG3[i]<-entropy(c(gammaG3[i],1-gammaG3[i]))-entropy(c(gamma_oldG3[i],1-gamma_oldG3[i]))
  #  }  
  #  appr_errorG3<-max(abs(entropy_changeG3))
  #}
  #a<-toc()
  #timeG3[j]<-a$toc-a$tic
  #errorG3[j]<-sum((theta-muG3*gammaG3)^2)  
  #fdrG3[j]<-sum((theta==0)*(gammaG3>0.5))/max(1,sum(gammaG3>0.5))
  #tprG3[j]<-sum((theta!=0)*(gammaG3>0.5))/sum(theta!=0)
  
  
  
  
}


# Computing the mean and standard deviation of the l2-error, FDR, TPR and runtime over 'repeat_number' runs:

errorLavg<-mean(sqrt(errorL))
errorLsd<-sd(sqrt(errorL))
errorL2avg<-mean(sqrt(errorL2))
errorL2sd<-sd(sqrt(errorL2))
errorG1avg<-mean(sqrt(errorG))
errorG1sd<-sd(sqrt(errorG))
errorG2avg<-mean(sqrt(errorG2))
errorG2sd<-sd(sqrt(errorG2))
errorG3avg<-mean(sqrt(errorG3))
errorG3sd<-sd(sqrt(errorG3))

errorLavg
errorLsd
errorL2avg
errorL2sd
errorG1avg
errorG1sd
errorG2avg
errorG2sd
errorG3avg
errorG3sd




fdrLavg<-mean(fdrL)
fdrLsd<-sd(fdrL)
fdrL2avg<-mean(fdrL2)
fdrL2sd<-sd(fdrL2)
fdrG1avg<-mean(fdrG)
fdrG1sd<-sd(fdrG)
fdrG2avg<-mean(fdrG2)
fdrG2sd<-sd(fdrG2)
fdrG3avg<-mean(fdrG3)
fdrG3sd<-sd(fdrG3)

fdrLavg
fdrLsd
fdrL2avg
fdrL2sd
fdrG1avg
fdrG1sd
fdrG2avg
fdrG2sd
fdrG3avg
fdrG3sd

tprLavg<-mean(tprL)
tprLsd<-sd(tprL)
tprL2avg<-mean(tprL2)
tprL2sd<-sd(tprL2)
tprG1avg<-mean(tprG)
tprG1sd<-sd(tprG)
tprG2avg<-mean(tprG2)
tprG2sd<-sd(tprG2)
tprG3avg<-mean(tprG3)
tprG3sd<-sd(tprG3)

tprLavg
tprLsd
tprL2avg
tprL2sd
tprG1avg
tprG1sd
tprG2avg
tprG2sd
tprG3avg
tprG3sd

timeLavg<-mean(timeL)
timeLsd<-sd(timeL)
timeL2avg<-mean(timeL2)
timeL2sd<-sd(timeL2)
timeG1avg<-mean(timeG)
timeGsd<-sd(timeG)
timeG2avg<-mean(timeG2)
timeG2sd<-sd(timeG2)
timeG3avg<-mean(timeG3)
timeG3sd<-sd(timeG3)



# Figures for the last run. Comparing the signal recovering property of the posterior means.

plot(c(1:p),theta , col=3, ylab="Signal value", xlab="Index",  ylim=c(-1,12.5), pch=20,cex.lab=1.5, cex.axis=1.5, cex.main=1.5, cex.sub=1.5)
#points(c(1:p),Y,col="grey",pch=20) # incase of Gaussian sequence model - Experiment 1
points(c(1:p),muG2*gammaG2,col="orange",pch=6)
points(c(1:p),muG*gammaG,col=2,pch=5)
points(c(1:p),muL2*gammaL2,col=6,pch=3)
points(c(1:p),muL*gammaL,col=4,pch=2)
points(c(1:p),theta,col=3,pch=20,cex=0.7)
legend("topright", col=c(3,4,6,2,"orange"),cex=1.2, lwd=2, legend=c("signal","Laplace P", "Laplace Q", "Gauss","Gauss (big var)"))


################### subroutins for the VB - Laplace algorithm:

fsigm<-function(i,sigm,mu){
  return(XTX[i,i]*sigm[i]^2/2+lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2))+lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i]))-log(sigm[i]))
}

fsigmhelp<-function(sigm_i){
  sigm_help<-sigm
  sigm_help[i]<- sigm_i
  return(fsigm(i,sigm_help,mu))
}
fmu<-function(i, sigm, mu,gamma){
  return( XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]+ XTX[i,i]*mu[i]^2/2-YTX[i]*mu[i] +lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2)) +lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])))
}
fmuhelp<-function(mu_i){
  mu_help<-mu
  mu_help[i]<- mu_i
  return(fmu(i,sigm,mu_help,gamma))
}
fgamma<-function(i, sigm, mu,gamma){
  help1<-log(a0/b0)+log(sqrt(pi)*sigm[i]*lambda/sqrt(2))
  help2<-YTX[i]*mu[i]- XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]-(mu[i]^2+sigm[i]^2)/2*XTX[i,i]-lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2)) -lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])) +1/2  
  return(inv.logit(help1+help2))
}

fgamma_eval<-function(i, sigm, mu,gamma,omega){
  help1<-log((2/pi)^(1/2)/(sigm[i]*lambda))-1/2+ log(b0/a0)
  help2<-XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]+(mu[i]^2+sigm[i]^2)/2*XTX[i,i]-YTX[i]*mu[i]+lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2))+lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])) 
  help3<-gamma[i]*(help1+help2)
  return(help3)
}

