##################################################################################################
#####  Comparing prioritised with lexicographical and  randomized  updating #######################
###################################################################################################
################################# Section 5.1 #####################################################
###################################################################################################




library(mvtnorm)
library(boot)
library(entropy)
require(graphics)
library(tictoc)
library(sparsevb)

############## initialisation ###############################


n<-100
p<-200
s<-20
value=10
theta<-rep(0,p)



# Location of non-zero signal coefficients:

#perm<-sample(c(1:p), s, replace = FALSE, prob = NULL)
#theta[perm]<-value
#theta[(p/2-s/2+1):(p/2+s/2)]<-value
theta[(p-s+1):p]<-value
#theta[1:s]<-value
theta<-t(theta)

########################################################################################
########### initialization for priority, lexicographical and random updating schemes
########################################################################################
repeat_number<-200
timeLp<-rep(0,repeat_number)
errorLp<-rep(0,repeat_number)
fdrLp<-rep(0,repeat_number)
tprLp<-rep(0,repeat_number)

timeLn<-rep(0,repeat_number)
errorLn<-rep(0,repeat_number)
fdrLn<-rep(0,repeat_number)
tprLn<-rep(0,repeat_number)

timeLr<-rep(0,repeat_number)
errorLr<-rep(0,repeat_number)
fdrLr<-rep(0,repeat_number)
tprLr<-rep(0,repeat_number)





for(iter in 1:repeat_number){
  print(iter)
  ######## data generation
  X<-matrix(rnorm(n*p, mean=0,sd=1),nrow=n,ncol=p)
  eps   =  rnorm(n,0,1)        # Noise epsilon
  Y     =  X %*% t(theta) +eps


  
  
  ############################ initialization of the  methods ##############

  
  ##############################################################################
  ################ Prioritized updating scheme #######################################
  #################################################################
  tic()
  vbL<-svb.fit(X,Y,tol = 10e-5, max_iter = 10^4, exact_math = TRUE, family ="linear")
  mu<-vbL$mu
  gamma<-vbL$gamma
  sigma<-vbL$sigma
  time0<-toc()
  timeLp[iter]<-time0$toc-time0$tic
  errorLp[iter]<-sqrt(sum((theta-t(mu*gamma))^2))
  fdrLp[iter]<-sum((theta==0)*(gamma>0.5))/max(sum(gamma>0.5),1)
  tprLp[iter]<-sum((theta!=0)*(gamma>0.5))/sum(theta!=0)
  
  # time, l2-error, FDr, TPR:
  
  postmean1<-mu*gamma
  
  ##############################################################################
  ################ Lexicographical updating scheme #######################################
  #################################################################
  
  
  tic()
  
  #vbL<-varselect(X,Y,tol = 10e-5, alpha=1, beta=p, prioritized_init = FALSE, update_order=(c(1:p)-1))
  vbL<-svb.fit(X,Y,family ="linear",tol = 10e-4, max_iter = 10^4, exact_math = TRUE,prioritized_init = FALSE, update_order=(c(1:p)-1))
  
  
  mu<-vbL$mu
  gamma<-vbL$gamma
  sigma<-vbL$sigma
  time0<-toc()
  
  # time, l2-erro, FDR, TPR:
  timeLn[iter]<-time0$toc-time0$tic
  errorLn[iter]<-sqrt(sum((theta-t(mu*gamma))^2))
  fdrLn[iter]<-sum((theta==0)*(gamma>0.5))/max(sum(gamma>0.5),1)
  tprLn[iter]<-sum((theta!=0)*(gamma>0.5))/sum(theta!=0)
  postmean2<-mu*gamma
  
  
  ##############################################################################
  ################ Randomized updating scheme #######################################
  #################################################################
  
  tic()
  perm<-sample(c(1:p), p, replace = FALSE, prob = NULL) # randomized updating scheme
  #vbL<-varselect(X,Y,tol = 10e-5, alpha=1, beta=p, prioritized_init = FALSE, update_order=perm-1)
  vbL<-svb.fit(X,Y,family ="linear",tol = 10e-4, max_iter = 10^4, exact_math = TRUE,prioritized_init = FALSE, update_order=(perm-1))
  
  mu<-vbL$mu
  gamma<-vbL$gamma
  sigma<-vbL$sigma
  time0<-toc()
  
  timeLr[iter]<-time0$toc-time0$tic
  errorLr[iter]<-sqrt(sum((theta-t(mu*gamma))^2))
  fdrLr[iter]<-sum((theta==0)*(gamma>0.5))/max(sum(gamma>0.5),1)
  tprLr[iter]<-sum((theta!=0)*(gamma>0.5))/sum(theta!=0)
  postmean3<-mu*gamma

}

# computing the mean and the standard deviations of all these indicatiors (l2-erro, FDR, TRP, runtime)
errorLpavg<-mean(errorLp)
errorLpsd<-sd(errorLp)
fdrLpavg<-mean(fdrLp)
fdrLpsd<-sd(fdrLp)
tprLpavg<-mean(tprLp)
tprLpsd<-sd(tprLp)
timeLpavg<-mean(timeLp)
timeLpsd<-sd(timeLp)


errorLnavg<-mean(errorLn)
errorLnsd<-sd(errorLn)
fdrLnavg<-mean(fdrLn)
fdrLnsd<-sd(fdrLn)
tprLnavg<-mean(tprLn)
tprLnsd<-sd(tprLn)
timeLnavg<-mean(timeLn)
timeLnsd<-sd(timeLn)


errorLravg<-mean(errorLr)
errorLrsd<-sd(errorLr)
fdrLravg<-mean(fdrLr)
fdrLrsd<-sd(fdrLr)
tprLravg<-mean(tprLr)
tprLrsd<-sd(tprLr)
timeLravg<-mean(timeLr)
timeLrsd<-sd(timeLr)


errorLpavg
errorLpsd
errorLnavg
errorLnsd
errorLravg
errorLrsd

fdrLpavg
fdrLpsd
fdrLnavg
fdrLnsd
fdrLravg
fdrLrsd

tprLpavg
tprLpsd
tprLnavg
tprLnsd
tprLravg
tprLrsd


timeLpavg
timeLpsd
timeLnavg
timeLnsd
timeLravg
timeLrsd

# Plotting the pictures:
plot(c(1:p),theta , col="white", ylab="Signal value", xlab="Index", ylim=c(-8,20), pch=20,cex.lab=1.5, cex.axis=1.5, cex.main=1.5, cex.sub=1.5)
points(c(1:p),postmean1,col=4,pch=2)
points(c(1:p),postmean2,col=2,pch=5)
points(c(1:p),postmean3,col="orange",pch=1)
points(c(1:p),theta,col=3,pch=20,cex = 0.7)
legend("topright", col=c(3,4,2,"orange"),cex=1.4, lwd=2, legend=c("signal","vb-Laplace (priority)","vb-Laplace (lexicographical)", "vb-Laplace (random)"))


