#####################################################################################################
#############################   Comparing Bayesian variable selection methods #################################
###################################################################################################
################################# Sections 5.2 and A.5 ##################################################################
###################################################################################################


library(mvtnorm)
library(boot)
library(entropy)
require(graphics)
library(tictoc)
library(matlib)
library(EMVS)
library(spikeslab)
library(varbvs)
library(selectiveInference)
require(MASS)
library(SSLASSO)
library(horseshoe)
library(lars)
library(matrixStats)



############################################################################################################################################################################################################################################################
##############################################################################################################################

# Experiment 1:
#n<-100
#p<-n*4
#s<-20
#value<-log(n)
#theta<-rep(0,p)
#theta[(p-s+1):p]<- value
#sigma<-5


# Experiment 2:

#p <- 1000
#n <- 100
#theta <- c(1, 2, 3, rep(0, p-3))
#sigma<-1


# Experiment 3:
n<-200
p<-n*4
s<-5
value<-runif(s,-5,5)
theta<-rep(0,p)
tau<-1
theta[((p-s)/2):((p+s)/2)]<- value
sigma<-0.2


# Experiment 4:
#n<-100
#p<-n*4
#s<-10
#value<-runif(s,-3,3)
#theta<-rep(0,p)
#theta[1:s]<- value
#sigma<-0.2


# Experiment 5:
#n<-200
#p<-n
#s<-20
#value=runif(s,-10,10)
#theta<-rep(0,p)
#theta[(p-s+1):p]<- value
#sigma<-200

# Experiment 6:
#n<-100
#p<-n*4
#s<-20
#value<-log(n)
#theta<-rep(0,p)
#theta[(p-s+1):p]<- value
#sigma<-2



# Experiment 7:
#n<-100
#p<-n*4
#s<-20
#value<-2*log(n)
#theta<-rep(0,p)
#theta[(p-s+1):p]<- value
#sigma<-5


# Experiment 8:
#n<-200
#p<-300
#s<-15
#value<-runif(s,-2,2)
#theta<-rep(0,p)
#tau<-1
#theta[(p-s+1):p]<- value
#sigma<-2



v_number<-4
lambda_number<-5
repeat_number<-5


errorvar<-rep(0,repeat_number)
fdrvar<-rep(0,repeat_number)
tprvar<-rep(0,repeat_number)
timevar<-rep(0,repeat_number)

errorvar_p<-rep(0,repeat_number)
fdrvar_p<-rep(0,repeat_number)
tprvar_p<-rep(0,repeat_number)
timevar_p<-rep(0,repeat_number)


errorhorse<-rep(0,repeat_number)
errorhorse_p<-rep(0,repeat_number)
timehorse<-rep(0,repeat_number)
timehorse_p<-rep(0,repeat_number)


errorSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
fdrSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
tprSSL<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
timeSSL<-rep(0,repeat_number)


errorSSL_p<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
fdrSSL_p<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
tprSSL_p<-matrix(rep(0,lambda_number*repeat_number),nrow=repeat_number )
timeSSL_p<-rep(0,repeat_number)



for (j in 1:repeat_number){
print(j)
  # Generating the data and the design matrix:
  eps   =  rnorm(n,0,1)  
  X<-matrix(rnorm(n*p, mean=0,sd=1),nrow=n,ncol=p)
  #X<-standard(X)
  Y =  X %*% theta +eps*sigma

  
  
  ########################### varbvs ###########################
  
  tic()
  #initialisation:
  XTX= t(X) %*% X
  YTX<-t(X)%*%Y
  mu0<-solve(XTX+diag(p)*0.01)%*%YTX
  perm<-order(-abs(mu0))
  #perm<-c(p:1)
  
  help<-varbvs(X=X, Z=NULL,y=Y, family = "gaussian",  mu=mu0 ,verbose = TRUE, tol = 1e-3, maxiter = 10000)
  thetaM<-help$beta
  pip<-help$pip
  time0<-toc()
  timevar[j]<-time0$toc-time0$tic
  errorvar[j]<-sqrt(sum((theta-t(thetaM))^2))
  fdrvar[j]<-sum((theta==0)*(pip>0.5))/max(1,sum(pip>0.5))
  tprvar[j]<-sum((theta!=0)*(pip>0.5))/sum(theta!=0)
  
  
  
  
  ########################### varbvs - priority ###########################
  
  tic()
  #initialisation:
  X_p<-X[,perm]
  
  XTX_p= t(X_p) %*% X_p
  YTX_p<-t(X_p)%*%Y
  mu0_p<-solve(XTX_p+diag(p))%*%YTX_p
  
  help<-varbvs(X=X_p, Z=NULL,y=Y, family = "gaussian",  mu=mu0_p ,verbose = TRUE, tol = 1e-3, maxiter = 10000)
  thetaM_p<-help$beta
  pip_p<-help$pip
  time0<-toc()
  timevar_p[j]<-time0$toc-time0$tic
  errorvar_p[j]<-sqrt(sum((theta[perm]-t(thetaM_p))^2))
  fdrvar_p[j]<-sum((theta==0)*(pip>0.5))/max(1,sum(pip_p>0.5))
  tprvar_p[j]<-sum((theta!=0)*(pip_p>0.5))/sum(theta!=0)
  
  
  ########################### horseshoe###########################
  tic()
 help<- horseshoe(Y, X, method.tau = "truncatedCauchy",
            tau = 1, method.sigma = "Jeffreys", Sigma2 = 1,
            burn = 1000, nmc = 5000, thin = 1, alpha = 0.05)
  thetahorse<-help$BetaHat
  
 time0<-toc()
 timehorse[j]<-time0$toc-time0$tic
 errorhorse[j]<-sqrt(sum((theta-thetahorse)^2))
 
 ########################### horseshoe - priority ###########################
 tic()
 help_p<- horseshoe(Y, X_p, method.tau = "truncatedCauchy",
                    tau = 1, method.sigma = "Jeffreys", Sigma2 = 1,
                    burn = 1000, nmc = 5000, thin = 1, alpha = 0.05)
 thetahorse_p<-help_p$BetaHat
 
 time0<-toc()
 timehorse_p[j]<-time0$toc-time0$tic
 errorhorse_p[j]<-sqrt(sum((theta[perm]-thetahorse_p)^2))
 
 
  ########################### SSLASSO ###########################
  tic()
  # initialisation:
  lambda1<-0.01
  lambda0<-seq(lambda1,p,length=lambda_number)
  # calling the SSLASSO routine
  obj<-SSLASSO(X, Y, variance = "unknown",lambda1 = lambda1, lambda0 = lambda0,a=1,b=p)
  thetaSSL<-obj$beta
  pipSSL<-obj$select
  time0<-toc()
  timeSSL[j]<-time0$toc-time0$tic
  
  for (i in 1:lambda_number){
    errorSSL[j,i]<-sqrt(sum((theta-thetaSSL[,i])^2))
    fdrSSL[j,i]<-sum((theta==0)*(pipSSL[,i]>0.5))/max(sum(pipSSL[,i]>0.5),1)
    tprSSL[j,i]<-sum((theta!=0)*(pipSSL[,i]>0.5))/sum(theta!=0)
  } 
  
  ########################### SSLASSO - priority ###########################
  tic()
  # initialisation:
  lambda1<-0.01
  lambda0<-seq(lambda1,p,length=lambda_number)
  # calling the SSLASSO routine
  obj_p<-SSLASSO(X_p, Y, variance = "unknown",lambda1 = lambda1, lambda0 = lambda0,a=1,b=p)
  thetaSSL_p<-obj_p$beta
  pipSSL_p<-obj_p$select
  time0<-toc()
  timeSSL_p[j]<-time0$toc-time0$tic
  
  for (i in 1:lambda_number){
    errorSSL_p[j,i]<-sqrt(sum((theta[perm]-thetaSSL_p[,i])^2))
    fdrSSL_p[j,i]<-sum((theta==0)*(pipSSL_p[,i]>0.5))/max(sum(pipSSL_p[,i]>0.5),1)
    tprSSL_p[j,i]<-sum((theta!=0)*(pipSSL_p[,i]>0.5))/sum(theta!=0)
  } 
  
   
   
}


# Computing the mean and standard deviation of the above indicatior over 'repeat_number' runs:


errorvaravg<-mean(errorvar)
errorvarsd<-sd(errorvar)
errorvaravg_p<-mean(errorvar_p)
errorvarsd_p<-sd(errorvar_p)

errorhorseavg<-mean(errorhorse)
errorhorsesd<-sd(errorhorse)
errorhorseavg_p<-mean(errorhorse_p)
errorhorsesd_p<-sd(errorhorse_p)

error_meanSSL<-colMeans(errorSSL)
error_sdSSL<-colSds(errorSSL)
error_meanSSL_p<-colMeans(errorSSL_p)
error_sdSSL_p<-colSds(errorSSL_p)


errorvaravg
errorvarsd
errorvaravg_p
errorvarsd_p
errorhorseavg
errorhorsesd
errorhorseavg_p
errorhorsesd_p
error_meanSSL
error_sdSSL
error_meanSSL_p
error_sdSSL_p


fdrvaravg<-mean(fdrvar)
fdrvarsd<-sd(fdrvar)
fdrEMVSavg<-colMeans(fdrEMVS)
fdrEMVSsd<-colSds(fdrEMVS)
fdrSSLavg<-colMeans(fdrSSL)
fdrSSLsd<-colSds(fdrSSL)
fdrEBavg<-mean(fdrEB)
fdrEBsd<-sd(fdrEB)

fdrLavg
fdrLsd
fdrvaravg
fdrvarsd
fdrEMVSavg
fdrEMVSsd
fdrSSLavg
fdrSSLsd
fdrEBavg
fdrEBsd

tprLavg<-mean(tprL)
tprLsd<-sd(tprL)
tprvaravg<-mean(tprvar)
tprvarsd<-sd(tprvar)
tprEMVSavg<-colMeans(tprEMVS)
tprEMVSsd<-colSds(tprEMVS)
tprSSLavg<-colMeans(tprSSL)
tprSSLsd<-colSds(tprSSL)
tprEBavg<-mean(tprEB)
tprEBsd<-sd(tprEB)

tprLavg
tprLsd
tprvaravg
tprvarsd
tprEMVSavg
tprEMVSsd
tprSSLavg
tprSSLsd
tprEBavg
tprEBsd

timeLavg<-mean(timeL)
timeLsd<-sd(timeL)
timevaravg<-mean(timevar)
timevarsd<-sd(timevar)
timeEMVSavg<-mean(timeEMVS)
timeEMVSsd<-sd(timeEMVS)
timeSSLavg<-mean(timeSSL)
timeSSLsd<-sd(timeSSL)
timeEBavg<-mean(timeEB)
timeEBsd<-sd(timeEB)

timeLavg
timeLsd
timevaravg
timevarsd
timeEMVSavg
timeEMVSsd
timeSSLavg
timeSSLsd
timeEBavg
timeEBsd





########################### Pictures - for the last run ###########################

plot(c(1:p),theta , col=3, ylab="Signal value", xlab="Index",  ylim=c(-1*max(theta),2*max(theta)), pch=20,cex.lab=1.5, cex.axis=1.5, cex.main=1.5, cex.sub=1.5)

points(c(1:p),theta,col=3,pch=20)
points(c(1:p),mu*gamma,col=4,pch=20)
points(c(1:p),thetaM, col=2,pch=20)
points(c(1:p),thetaEMVS, col=7,pch=20)
points(c(1:p),thetaSSL, col=5,pch=20)
points(c(1:p),postmean_MCMC, col=8,pch=20)
#legend("topright", col=c(3,4,2,6,7,5,8),cex=1.2, lwd=2, legend=c("signal", "vb - Laplace", "varbvs", "spikeslab","EMVS","SSLASSO", "Empirical Bayes"))








