################################################################################
############################# CV-error in real world data #################################
###################################################################################################
################################# Section A.1 ##################################################################
###################################################################################################


library(boot)
library(entropy)
library(spikeslab)
require(graphics)
library(tictoc)
library(matlib)
library(EMVS)
library(spikeslab)
library(varbvs)
library(selectiveInference)
require(MASS)
library(SSLASSO)
library(horseshoe)
library(lars)
library(R.matlab)
library(sparsevb)


# ozone dataset, centering, rescaling and adding an interceptparameter (except for EMVS the last one)
data(ozoneI, package = "spikeslab")
Yfull <- ozoneI[, 1]
Xfull <- ozoneI[, -1]
Xfull<-scale(Xfull, scale = TRUE, center=TRUE)
Xfull_EMVS<- Xfull
intercept<-rep(1,n)
Xfull <- cbind(Xfull,intercept )
Xfull<-data.matrix(Xfull,rownames.force = NA)
Xfull_EMVS<-data.matrix(Xfull_EMVS,rownames.force = NA)
n<-nrow(Xfull)
p<-ncol(Xfull) 


#initialization:
number_of_methods<-6
cverror<-rep(0,5)
lambda_number<-100
cverrorSSL<-rep(0,lambda_number)
fold<-10
timeL<-0
timevar<-0
timeSS<-0
timeSSL<-0
timeEB<-0
#estimation of sigma
hatsigma<-0
while(hatsigma<0.01|hatsigma>100 | hatsigma=="NaN"){
  hatsigma<-estimateSigma(x=Xfull,y=Yfull )$sigmahat
}


##############################################################################
 
for (iter in 1:fold){  #CV
  X<-Xfull[-(((iter-1)*n/fold+1):(iter*n/fold)), ]
  Y<-Yfull[-(((iter-1)*n/fold+1):(iter*n/fold))]
  Xfold<-Xfull[(((iter-1)*n/fold+1):(iter*n/fold)), ]
  Yfold<-Yfull[(((iter-1)*n/fold+1):(iter*n/fold))]
  X_EMVS<-Xfull_EMVS[-(((iter-1)*n/fold+1):(iter*n/fold)), ]
  Xfold_EMVS<-Xfull_EMVS[(((iter-1)*n/fold+1):(iter*n/fold)), ]
  

  
  ################ VB - Laplace slab #############
  tic() 
  vbL<-svb.fit(X,Y,lambda=lambda,tol = 10e-4, max_iter = 10^4, exact_math = TRUE, family ="linear")
  
  mu<-vbL$mu
  sigma<-vbL$sigma
  gamma<-vbL$gamma
  time0<-toc()
  
  #time, CV-error sum:
  timeL<-timeL+time0$toc-time0$tic
  cverror[1]<-cverror[1]+sqrt(sum((Yfold- Xfold%*%(gamma*mu) )^2)) 
  
  
  
  ########################## ########################## ########################## 
  ##########################  varbvs   #########################
  ########################## ########################## ########################## 
  tic()
  #calling varbvs subroutine
  fvarbvs<-varbvs(X=X, Z=NULL,y=Y, family = "gaussian", verbose = TRUE, tol = 1e-4, maxiter = 10000)
  pvar<-fvarbvs$alpha[,20]
  postmean<-fvarbvs$mu[,20]
  time0<-toc()
  #time, CV-error sum:
  timevar<-timevar+time0$toc-time0$tic
  cverror[2]<-cverror[2]+sqrt(sum((Yfold- Xfold%*%postmean )^2)) 
  
  
   

  
  ###############################################################
  ##################### EMVS #############################
  ###############################################################

  
  tic()
  #initialization:
  XTX_emvs= t(X_EMVS) %*% X_EMVS
  YTX_emvs<-t(X_EMVS)%*%Y
  mu0_emvs<-solve(XTX_emvs+diag(p-1))%*%YTX_emvs
  v0 = 0.1
  v1 = 1000
  a = 1
  b = p-1
  epsilon = 10^{-3}
  beta_init<-mu0_emvs
  # callingthe EMVS routine:
  obj<-EMVS(Y, X_EMVS,v0=v0, v1=v1,type = c("betabinomial"), independent = TRUE,
            beta_init, epsilon=epsilon,a=a,b=b)
  
  pEMVS<-obj$prob_inclusion
  postmeanEMVS<-obj$betas
  time0<-toc()
  #time, CV-error sum:
  timeEMVS<-time0$toc-time0$tic
  cverror[3]<-cverror[3]+sqrt(sum((Yfold- Xfold_EMVS%*%t(postmeanEMVS) )^2)) 
  
  
  ##################################################################
  ########################### SSLASSO ###########################
  ##################################################################

  tic() 
  ### initialization  ###
  lambda1<-0.01
  lambda0<-seq(lambda1,p,length=lambda_number)
  # calling SSLASSO function
  obj<-SSLASSO(X, Y, variance = "unknown",lambda1 = lambda1, lambda0 = lambda0,a=1,b=p)
  postmeanSSL<-obj$beta
  pSSL<-obj$select
  time0<-toc()
  #time, CV-error sum:
  timeSSL<-timeSSL+time0$toc-time0$tic
  for (i in 1:lambda_number){
    cverrorSSL[i]<-cverrorSSL[i]+sqrt(sum((Yfold- Xfold%*%postmeanSSL[,i] )^2))
    
  }
  
  ############################################################################  
########################### ebreg ############################################
#################################################################################
##### Didn't run due to problems with Cholesky-decomposition
  #tic()
  ##initialisation
  #sample.beta <- TRUE
  #alpha <- 0.99
  #gam <- 0.001
  #M<-5000
  #log.f <- function(x) dcomplex(x, n, p, 0.05, 1)
  ## estimating the variance
  #hatsigma<-0
  #while(hatsigma<0.01|hatsigma>100 | hatsigma=="NaN"){
  #  hatsigma<-estimateSigma(x=X,y=Y)$sigmahat
  #}
  # calling the ebreg.R routine
  #o <- ebreg(Y, X, sig2=hatsigma, log.f, alpha, gam, M, sample.beta)
  #sample_beta<-o$beta
  #postmean_MCMC<-colMeans(sample_beta, na.rm = FALSE, dims = 1)
  #time0<-toc()
  #timeEB<-timeEB+time0$toc-time0$tic
  #cverror[5]<-cverror[5]+sum((Yfold- Xfold%*%postmean_MCMC )^2) 

}


## divide cv error by the number of folds (we take the average cv-error)
cverrorSSL<-cverrorSSL/fold
cverror<-cverror/fold
cverrorSSL
cverror












##############################################################################################################################
########################################## subroutins     3##########################################################################
##############################################################################################################################

############################  For the VB Laplace method ############################ 
fsigm<-function(i,sigm,mu){
  return(XTX[i,i]*sigm[i]^2/2+lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2))+lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i]))-log(sigm[i]))
}

fsigmhelp<-function(sigm_i){
  sigm_help<-sigm
  sigm_help[i]<- sigm_i
  return(fsigm(i,sigm_help,mu))
}
fmu<-function(i, sigm, mu,gamma){
  return( XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]+ XTX[i,i]*mu[i]^2/2-YTX[i]*mu[i] +lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2)) +lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])))
}
fmuhelp<-function(mu_i){
  mu_help<-mu
  mu_help[i]<- mu_i
  return(fmu(i,sigm,mu_help,gamma))
}
fgamma<-function(i, sigm, mu,gamma){
  help1<-log(a0/b0)+log(sqrt(pi)*sigm[i]*lambda/sqrt(2))
  help2<-YTX[i]*mu[i]- XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]-(mu[i]^2+sigm[i]^2)/2*XTX[i,i]-lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2)) -lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])) +1/2  
  return(inv.logit(help1+help2))
}


############################ For the ebreg routine ############################ ############################ 



ebreg <- function(y, X, sig2=NULL, log.f, alpha, gam, M, sample.beta=FALSE) {
  
  n <- nrow(X)
  p <- ncol(X)
  if(!exists("lars")) library(lars)
  o.lars <- lars(X, y, normalize=FALSE, intercept=FALSE, use.Gram=FALSE)
  cv <- cv.lars(X, y, plot.it=FALSE, se=FALSE, normalize=FALSE, intercept=FALSE, use.Gram=FALSE)
  b.lasso <- coef(o.lars, s=cv$index[which.min(cv$cv)], mode="fraction")
  S <- as.numeric(b.lasso != 0)
  if(is.null(sig2)) {
    
    z <- as.numeric(y - X[, S > 0] %*% b.lasso[S > 0])
    sig2 <- sum(z**2) / max(n - sum(S), 1)
    
  }
  v <- 1 + alpha / gam / sig2
  sq.v <- sqrt(v)
  B <- round(0.2 * M)
  bb <- if(sample.beta) matrix(0, nrow=B + M, ncol=p) else NULL
  SS <- matrix(0, nrow=B + M, ncol=p)
  ss <- numeric(B + M)
  ii <- integer(B + M)
  SS[1,] <- S
  ss[1] <- s <- sum(S)
  ii[1] <- i <- iuS <- nuS <- 1
  out <- list()
  out[[1]] <- get.lm.stuff(S, y, X, sample.beta)
  lprior <- log.f(0:n) - lchoose(p, 0:n)
  lpost <- lprior[s] - alpha * out[[1]]$sse / 2 / sig2 - s * log(v) / 2
  for(m in 1:(B + M)) {
    
    S.new <- rprop(S, n)
    s.new <- sum(S.new)
    i.new <- compare.to.rows(as.matrix(SS[iuS,]), S.new)
    if(i.new == 0) o.new <- get.lm.stuff(S.new, y, X, sample.beta) else o.new <- out[[i.new]]
    lpost.new <- lprior[s.new] - alpha * o.new$sse / 2 / sig2 - s.new * log(v) / 2
    if(runif(1) <= exp(lpost.new - lpost)) {
      
      S <- S.new
      s <- s.new
      lpost <- lpost.new
      if(i.new == 0) {
        
        nuS <- nuS + 1
        i <- nuS
        iuS <- c(iuS, m)
        out[[nuS]] <- o.new
        
      } else i <- i.new
      
    }
    SS[m,] <- S
    ss[m] <- s
    ii[m] <- i
    if(sample.beta) {
      
      b <- numeric(p)
      b[S > 0] <- out[[i]]$b.hat + sq.v * out[[i]]$U %*% rnorm(s)
      bb[m,] <- b
      
    }
    
  }
  if(sample.beta) bb[-(1:B),]
  return(list(beta=bb, S=SS[-(1:B),], s=ss[-(1:B)], i=ii[-(1:B)]))
  
}



# Auxiliary functions called by 'ebreg'


dcomplex <- function(x, n, p, a, b, log=TRUE) {
  
  o <- -x * (log(b) + a * log(p)) + log(x <= n)
  if(!log) o <- exp(o)
  return(o)
  
}



rprop <- function(S, n) {
  
  s <- sum(S)
  if(s == n) { S[sample(which(S == 1), 1)] <- 0 }
  else if(s  == 1) { S[sample(which(S == 0), 1)] <- 1 }
  else {
    
    if(runif(1) <= 0.5) S[sample(which(S == 1), 1)] <- 0
    else S[sample(which(S == 0), 1)] <- 1
    
  }
  return(S)
  
}


compare.to.rows <- function(SS, S) {
  
  h <- function(v) sum(abs(v - S))
  o <- apply(SS, 1, h)
  if(all(o > 0)) return(0) else return(which(o == 0)[1])
  
}


get.lm.stuff <- function(S, y, X, sample.beta) {
  
  if(!exists("ginv")) library(MASS)
  X.S <- as.matrix(X[, S > 0])
  o <- lm(y ~ X.S - 1)
  sse <- sum(o$residuals**2)
  if(sample.beta) {
    
    b.hat <- o$coefficients
    XtX.S <- t(X.S) %*% X.S
    V <- ginv(XtX.S)
    U <- chol(V)
    
  } else {
    
    U <- NULL
    b.hat <- NULL
    
  }
  return(list(sse=sse, b.hat=b.hat, U=U))
  
}




