library(boot)
library(entropy)
library(spikeslab)
require(graphics)
library(tictoc)
library(matlib)
library(EMVS)
library(spikeslab)
library(varbvs)
library(selectiveInference)
require(MASS)
library(SSLASSO)
library(horseshoe)
library(lars)
library(R.matlab)
library(varpack)

# ozone dataset
data(ozoneI, package = "spikeslab")
Y <- ozoneI[, 1]
X <- ozoneI[, -1]
X<-scale(X, scale = TRUE, center=TRUE)
X<-data.matrix(X)
X_EMVS<- X # in case of the EMVS method we do not add an intercept parameter
intercept<-rep(1,n) 
X <- cbind(X,intercept )
n<-nrow(X)
p<-ncol(X) 



number_of_methods<-6
fold<-10
timeL<-0
timevar<-0
timeSS<-0
timeSSL<-0
timeEB<-0
lambda_number<-100
modelSSL<-rep(0,lambda_number)

 
 
################################################################################################
  ################ Component-wise variational Bayes algorithm - Laplace slab #############
################################################################################################
  tic()
  #estimation of unknown variance:
  hatsigma<-0
  while(hatsigma<0.01|hatsigma>100 | hatsigma=="NaN"){
    hatsigma<-estimateSigma(x=X,y=Y)$sigmahat
  }
  Xrescaled<-X/hatsigma
  Yrescaled<-Y/hatsigma
  vbL<-varselect(X=Xrescaled,Y=Yrescaled, beta=p)
  mu<-vbL$mu
  sigma<-vbL$sigma
  gamma<-vbL$gamma
  time0<-toc()
  
  # runtime and model size:
  timeL<-time0$toc-time0$tic
  modelL<-sum(gamma>0.5)
  
  ########################################################################################
  ##########################  varbvs   #########################
  ########################################################################################
  tic()
  XTX= t(X) %*% X
  YTX<-t(X)%*%Y
  mu0<-solve(XTX+diag(p))%*%YTX
  help<-varbvs(X=X, Z=NULL,y=Y, family = "gaussian", mu=mu0 ,verbose = TRUE, tol = 1e-5, maxiter = 10000)
  pvar<-help$alpha[,20]
  postmean<-help$mu[,20]
  time0<-toc()
  timevar<-time0$toc-time0$tic
  modelvar<-sum(pvar>0.5)
  modelvar
  
  
  
  ############################################################################
  ##################### EMVS #############################
  ############################################################################
  
  tic()
  #initialization:
  XTX_emvs= t(X_EMVS) %*% X_EMVS
  YTX_emvs<-t(X_EMVS)%*%Y
  mu0_emvs<-solve(XTX_emvs+diag(p-1))%*%YTX_emvs
  v0 = 0.1
  v1 = 1000
  a = 1
  b = p-1
  epsilon = 10^{-3}
  beta_init<-mu0_emvs
  # callingthe EMVS routine:
  obj<-EMVS(Y, X_EMVS,v0=v0, v1=v1,type = c("betabinomial"), independent = TRUE,
            beta_init, epsilon=epsilon,a=a,b=b)
  
  pEMVS<-obj$prob_inclusion
  postmeanEMVS<-obj$betas
  time0<-toc()
  
  # time and model size:
  timeEMVS<-time0$toc-time0$tic
  modelEMVS<-sum(pEMVS>0.5)
  
  
  ########################################################################
  ########################### SSLASSO ###########################
  ########################################################################
  # initialization:
  tic()
  lambda1<-0.01
  lambda0<-seq(lambda1,p,length=lambda_number)
  #calling SSLASSO subroutine:
  obj<-SSLASSO(X, Y, variance = "unknown",lambda1 = lambda1, lambda0 = lambda0,a=1,b=p)
  postmeanSSL<-obj$beta
  pSSL<-obj$select
  time0<-toc()
  #time and model size for different hyper-parameters:
  timeSSL<-time0$toc-time0$tic
  for (i in 1:lambda_number){
    modelSSL[i]<-sum(pSSL[,i]>0.5)
  }
  
  ########################################################################
  ########################### ebreg algorithm - Ryan Martin's code ###########################
  ########################################################################
  # Results in an error message: problem with Cholesky decomposition step due to colinearity.
  #tic()
  #sample.beta <- TRUE
  #alpha <- 0.99
  #gam <- 0.001
  #M<-5000
  #log.f <- function(x) dcomplex(x, n, p, 0.05, 1)
  #hatsigma<-0
  #while(hatsigma<0.01|hatsigma>100 | hatsigma=="NaN"){
  #  hatsigma<-estimateSigma(x=X,y=Y)$sigmahat
  #}
  #o <- ebreg(Y, X, sig2=hatsigma, log.f, alpha, gam, M, sample.beta)
  #sample_beta<-o$beta
  #postmean_MCMC<-colMeans(sample_beta, na.rm = FALSE, dims = 1)
  #time0<-toc()
  #timeEB<-time0$toc-time0$tic
  #sampleS<-o$S
  #postmean_S<-colMeans(sampleS, na.rm = FALSE, dims = 1)
  #modelebreg<-sum(postmean_S>0.5)
  #modelebreg



###################  Figures  ###########################





plot(c(1:(p-1)), pvar[1:134], col=2, ylab="posterior inclusion probability", xlab="Covariate index",  ylim=c(-0.2,1.8), pch=5,cex.lab=1.5, cex.axis=1.5, cex.main=1.5, cex.sub=1.5, main="ozone")
points(c(1:(p-1)),pEMVS, col=6,pch=20,cex=1.5)
points(c(1:(p-1)),pSSL[1:134,95], col=3,pch=5,cex=1.5)
points(c(1:(p-1)),pvar[1:134], col=2,pch=5)
points(c(1:(p-1)),gamma[1:134], col=4,pch=2)
#points(c(1:(p-1)),postmean_S[1:134], col="orange",pch=6)

legend("topright", col=c(4,2,6,3),cex=1.4, lwd=2, legend=c( "sparsevb", "varbvs","EMVS", "SSLASSO"))#, "ebreg"))


















###############   RUN THIS FIRST #################

##############################################################################################################################
########################################## subroutins     ##########################################################################
##############################################################################################################################
fsigm<-function(i,sigm,mu){
  return(XTX[i,i]*sigm[i]^2/2+lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2))+lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i]))-log(sigm[i]))
}

fsigmhelp<-function(sigm_i){
  sigm_help<-sigm
  sigm_help[i]<- sigm_i
  return(fsigm(i,sigm_help,mu))
}
fmu<-function(i, sigm, mu,gamma){
  return( XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]+ XTX[i,i]*mu[i]^2/2-YTX[i]*mu[i] +lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2)) +lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])))
}
fmuhelp<-function(mu_i){
  mu_help<-mu
  mu_help[i]<- mu_i
  return(fmu(i,sigm,mu_help,gamma))
}
fgamma<-function(i, sigm, mu,gamma){
  help1<-log(a0/b0)+log(sqrt(pi)*sigm[i]*lambda/sqrt(2))
  help2<-YTX[i]*mu[i]- XTX[-i,i]%*% (gamma*mu)[-i]*mu[i]-(mu[i]^2+sigm[i]^2)/2*XTX[i,i]-lambda*sigm[i]*sqrt(2/pi)*exp(-mu[i]^2/(2*sigm[i]^2)) -lambda*mu[i]*(1-2*pnorm(-mu[i]/sigm[i])) +1/2  
  return(inv.logit(help1+help2))
}






ebreg <- function(y, X, sig2=NULL, log.f, alpha, gam, M, sample.beta=FALSE) {
  
  n <- nrow(X)
  p <- ncol(X)
  if(!exists("lars")) library(lars)
  o.lars <- lars(X, y, normalize=FALSE, intercept=FALSE, use.Gram=FALSE)
  cv <- cv.lars(X, y, plot.it=FALSE, se=FALSE, normalize=FALSE, intercept=FALSE, use.Gram=FALSE)
  b.lasso <- coef(o.lars, s=cv$index[which.min(cv$cv)], mode="fraction")
  S <- as.numeric(b.lasso != 0)
  if(is.null(sig2)) {
    
    z <- as.numeric(y - X[, S > 0] %*% b.lasso[S > 0])
    sig2 <- sum(z**2) / max(n - sum(S), 1)
    
  }
  v <- 1 + alpha / gam / sig2
  sq.v <- sqrt(v)
  B <- round(0.2 * M)
  bb <- if(sample.beta) matrix(0, nrow=B + M, ncol=p) else NULL
  SS <- matrix(0, nrow=B + M, ncol=p)
  ss <- numeric(B + M)
  ii <- integer(B + M)
  SS[1,] <- S
  ss[1] <- s <- sum(S)
  ii[1] <- i <- iuS <- nuS <- 1
  out <- list()
  out[[1]] <- get.lm.stuff(S, y, X, sample.beta)
  lprior <- log.f(0:n) - lchoose(p, 0:n)
  lpost <- lprior[s] - alpha * out[[1]]$sse / 2 / sig2 - s * log(v) / 2
  for(m in 1:(B + M)) {
    
    S.new <- rprop(S, n)
    s.new <- sum(S.new)
    i.new <- compare.to.rows(as.matrix(SS[iuS,]), S.new)
    if(i.new == 0) o.new <- get.lm.stuff(S.new, y, X, sample.beta) else o.new <- out[[i.new]]
    lpost.new <- lprior[s.new] - alpha * o.new$sse / 2 / sig2 - s.new * log(v) / 2
    if(runif(1) <= exp(lpost.new - lpost)) {
      
      S <- S.new
      s <- s.new
      lpost <- lpost.new
      if(i.new == 0) {
        
        nuS <- nuS + 1
        i <- nuS
        iuS <- c(iuS, m)
        out[[nuS]] <- o.new
        
      } else i <- i.new
      
    }
    SS[m,] <- S
    ss[m] <- s
    ii[m] <- i
    if(sample.beta) {
      
      b <- numeric(p)
      b[S > 0] <- out[[i]]$b.hat + sq.v * out[[i]]$U %*% rnorm(s)
      bb[m,] <- b
      
    }
    
  }
  if(sample.beta) bb[-(1:B),]
  return(list(beta=bb, S=SS[-(1:B),], s=ss[-(1:B)], i=ii[-(1:B)]))
  
}



# Auxiliary functions called by 'ebreg'


dcomplex <- function(x, n, p, a, b, log=TRUE) {
  
  o <- -x * (log(b) + a * log(p)) + log(x <= n)
  if(!log) o <- exp(o)
  return(o)
  
}



rprop <- function(S, n) {
  
  s <- sum(S)
  if(s == n) { S[sample(which(S == 1), 1)] <- 0 }
  else if(s  == 1) { S[sample(which(S == 0), 1)] <- 1 }
  else {
    
    if(runif(1) <= 0.5) S[sample(which(S == 1), 1)] <- 0
    else S[sample(which(S == 0), 1)] <- 1
    
  }
  return(S)
  
}


compare.to.rows <- function(SS, S) {
  
  h <- function(v) sum(abs(v - S))
  o <- apply(SS, 1, h)
  if(all(o > 0)) return(0) else return(which(o == 0)[1])
  
}


get.lm.stuff <- function(S, y, X, sample.beta) {
  
  if(!exists("ginv")) library(MASS)
  X.S <- as.matrix(X[, S > 0])
  o <- lm(y ~ X.S - 1)
  sse <- sum(o$residuals**2)
  if(sample.beta) {
    
    b.hat <- o$coefficients
    XtX.S <- t(X.S) %*% X.S
    V <- ginv(XtX.S)
    U <- chol(V)
    
  } else {
    
    U <- NULL
    b.hat <- NULL
    
  }
  return(list(sse=sse, b.hat=b.hat, U=U))
  
}


